/*	Copyright (C) 2003-2015 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#include "fe/platform.h"

namespace fe
{

FE_DL_PUBLIC MutexInitFunction* Mutex::ms_fnInit=NULL;
FE_DL_PUBLIC MutexLockFunction* Mutex::ms_fnLock=NULL;
FE_DL_PUBLIC MutexFinishFunction* Mutex::ms_fnFinish=NULL;

FE_DL_PUBLIC MutexGuardInitFunction* Mutex::ms_fnGuardInit=NULL;
FE_DL_PUBLIC MutexGuardLockFunction* Mutex::ms_fnGuardLock=NULL;
FE_DL_PUBLIC MutexGuardFinishFunction* Mutex::ms_fnGuardFinish=NULL;

FE_DL_PUBLIC MutexConditionInitFunction* Mutex::ms_fnConditionInit=NULL;
FE_DL_PUBLIC MutexConditionWaitFunction* Mutex::ms_fnConditionWait=NULL;
FE_DL_PUBLIC MutexConditionNotifyFunction* Mutex::ms_fnConditionNotify=NULL;
FE_DL_PUBLIC MutexConditionFinishFunction* Mutex::ms_fnConditionFinish=NULL;

FE_DL_PUBLIC String Mutex::ms_support;
FE_DL_PUBLIC BWORD Mutex::ms_checked=FALSE;

Mutex::Mutex(void):
	m_recursive(FALSE),
	m_initializing(FALSE),
	m_pRawMutex(NULL)
#if FE_MTX_COUNT
	,m_fakeCount(0),
	m_lockCount(0)
#endif
{
	//* defer init
}

Mutex::~Mutex(void)
{
#if FE_MTX_COUNT
	FEASSERT(m_fakeCount==0);
	FEASSERT(m_lockCount==0);
#endif

	if(m_pRawMutex)
	{
		if(ms_fnFinish)
		{
			ms_fnFinish(m_recursive,m_pRawMutex);
		}
		m_pRawMutex=NULL;
	}
}

void* Mutex::raw(void)
{
	if(!m_pRawMutex && !m_initializing)
	{
//		confirm();

		if(ms_fnInit)
		{
			m_initializing=TRUE;
			m_pRawMutex=ms_fnInit(m_recursive);
			m_initializing=FALSE;
		}
	}

#if FE_CODEGEN<=FE_DEBUG
	if(!m_pRawMutex)
	{
		fprintf(stderr,"Mutex::raw NULL mutex\n");
	}
#endif

//	FEASSERT(m_pRawMutex);
	return m_pRawMutex;
}

BWORD Mutex::lock(void)
{
#if FE_MTX_COUNT
	FEASSERT(m_fakeCount>=0);
	FEASSERT(m_lockCount>=0);
#endif

	if(!ms_fnLock || m_initializing)
	{
//		confirm();

		if(!ms_fnLock || m_initializing)
		{
#if FE_MTX_COUNT
			m_fakeCount++;
//			fprintf(stderr,"Mutex::lock   %p FAKE %d %d\n",
//					this,m_fakeCount,m_lockCount);
#endif
			return FALSE;
		}
	}

	const BWORD success=ms_fnLock(m_recursive,raw(),false,false,false);

#if FE_MTX_COUNT
	if(success)
	{
		m_lockCount++;
//		fprintf(stderr,"Mutex::lock   %p %p %d %d\n",
//				this,m_pRawMutex,m_fakeCount,m_lockCount);
	}
#endif

	return success;
}

BWORD Mutex::unlock(void)
{
	if(!ms_fnLock || m_initializing)
	{
#if FE_MTX_COUNT
		FEASSERT(m_fakeCount>0);
		FEASSERT(m_lockCount==0);
		m_fakeCount--;
//		fprintf(stderr,"Mutex::unlock %p FAKE %d %d\n",
//				this,m_fakeCount,m_lockCount);
#endif
		return FALSE;
	}

#if FE_MTX_COUNT
	FEASSERT(m_fakeCount==0);
	FEASSERT(m_lockCount>0);

	m_lockCount--;
//	fprintf(stderr,"Mutex::unlock %p %p %d %d\n",
//			this,m_pRawMutex,m_fakeCount,m_lockCount);
#endif

	return ms_fnLock(m_recursive,raw(),true,false,false);
}

BWORD Mutex::tryLock(void)
{
	if(!ms_fnLock || m_initializing)
	{
#if FE_MTX_COUNT
		m_fakeCount++;
#endif
		return FALSE;
	}

	if(ms_fnLock(m_recursive,raw(),false,true,false))
	{
#if FE_MTX_COUNT
		m_lockCount++;
//		fprintf(stderr,"Mutex::tryLock %p %d %d\n",
//				this,m_fakeCount,m_lockCount);
#endif
		return TRUE;
	}

	return FALSE;
}

BWORD Mutex::lockReadOnly(void)
{
	if(!ms_fnLock || m_initializing)
	{
#if FE_MTX_COUNT
		m_fakeCount++;
#endif
		return FALSE;
	}

#if FE_MTX_COUNT
	m_lockCount++;
//	fprintf(stderr,"Mutex::lockReadOnly %p\n",this);
#endif

	return ms_fnLock(m_recursive,raw(),false,false,true);
}

BWORD Mutex::unlockReadOnly(void)
{
	if(!ms_fnLock || m_initializing)
	{
#if FE_MTX_COUNT
		m_fakeCount--;
#endif
		return FALSE;
	}

#if FE_MTX_COUNT
	m_lockCount--;
//	fprintf(stderr,"Mutex::unlockReadOnly %p\n",this);
#endif
	return ms_fnLock(m_recursive,raw(),true,false,true);
}

BWORD Mutex::tryLockReadOnly(void)
{
	if(!ms_fnLock || m_initializing)
	{
#if FE_MTX_COUNT
		m_fakeCount++;
#endif
		return FALSE;
	}

	if(ms_fnLock(m_recursive,raw(),false,true,true))
	{
#if FE_MTX_COUNT
		m_lockCount++;
//		fprintf(stderr,"Mutex::tryLockReadOnly %p %d %d\n",
//				this,m_fakeCount,m_lockCount);
#endif
		return TRUE;
	}

	return FALSE;
}

Mutex::Guard::Guard(Mutex& a_rMutex,BWORD a_readOnly):
	m_pMutex(&a_rMutex),
	m_pRawGuard(NULL),
	m_readOnly(a_readOnly)
{
//	confirm();

	if(Mutex::ms_fnGuardInit)
	{
//		fprintf(stderr,"Mutex::Guard::Guard %p\n",this);
		void* pRawMutex=m_pMutex->raw();
		if(pRawMutex)
		{
			m_pRawGuard=Mutex::ms_fnGuardInit(
					m_pMutex->m_recursive,pRawMutex,m_readOnly);
			FEASSERT(m_pRawGuard);
		}
	}
}

Mutex::Guard::~Guard(void)
{
	if(Mutex::ms_fnGuardFinish && m_pRawGuard)
	{
//		fprintf(stderr,"Mutex::Guard::~Guard %p\n",this);
		Mutex::ms_fnGuardFinish(m_pMutex->m_recursive,m_pRawGuard,m_readOnly);
	}
}

BWORD Mutex::Guard::lock(void)
{
	if(Mutex::ms_fnGuardLock && m_pRawGuard)
	{
//		fprintf(stderr,"Mutex::Guard::lock %p\n",this);
		return Mutex::ms_fnGuardLock(
				m_pMutex->m_recursive,m_pRawGuard,FALSE,FALSE,m_readOnly);
	}
	return FALSE;
}

BWORD Mutex::Guard::unlock(void)
{
	if(Mutex::ms_fnGuardLock && m_pRawGuard)
	{
//		fprintf(stderr,"Mutex::Guard::unlock %p\n",this);
		return Mutex::ms_fnGuardLock(
				m_pMutex->m_recursive,m_pRawGuard,TRUE,FALSE,m_readOnly);
	}
	return FALSE;
}

BWORD Mutex::Guard::tryLock(void)
{
	if(Mutex::ms_fnGuardLock && m_pRawGuard)
	{
//		fprintf(stderr,"Mutex::Guard::tryLock %p\n",this);
		return Mutex::ms_fnGuardLock(
				m_pMutex->m_recursive,m_pRawGuard,FALSE,TRUE,m_readOnly);
	}
	return FALSE;
}

Mutex::Condition::Condition(void):
	m_pRawCondition(NULL)
{
	if(Mutex::ms_fnConditionInit)
	{
		m_pRawCondition=Mutex::ms_fnConditionInit();
		FEASSERT(m_pRawCondition);
	}
}

Mutex::Condition::~Condition(void)
{
	if(Mutex::ms_fnConditionFinish)
	{
		FEASSERT(m_pRawCondition);
		Mutex::ms_fnConditionFinish(m_pRawCondition);
	}
}

BWORD Mutex::Condition::wait(Guard& a_rGuard)
{
	if(Mutex::ms_fnConditionWait)
	{
		void* pRawGuard=a_rGuard.m_pRawGuard;
		if(pRawGuard)
		{
			FEASSERT(m_pRawCondition);
			return Mutex::ms_fnConditionWait(m_pRawCondition,
					a_rGuard.m_pMutex->m_recursive,pRawGuard,
					a_rGuard.m_readOnly);
		}
	}
	return FALSE;
}

BWORD Mutex::Condition::notifyOne(void)
{
	if(Mutex::ms_fnConditionNotify)
	{
		FEASSERT(m_pRawCondition);
		return Mutex::ms_fnConditionNotify(m_pRawCondition,FALSE);
	}
	return FALSE;
}

BWORD Mutex::Condition::notifyAll(void)
{
	if(Mutex::ms_fnConditionNotify)
	{
		FEASSERT(m_pRawCondition);
		return Mutex::ms_fnConditionNotify(m_pRawCondition,TRUE);
	}
	return FALSE;
}

#ifdef _OPENMP
extern "C"
{
bool omp_mutex_init(void);
}
#endif

//* static
BWORD Mutex::confirm(void)
{
	if(ms_fnInit)
	{
		return TRUE;
	}
	if(ms_checked)
	{
		return FALSE;
	}
	ms_checked=TRUE;

	const String verbose=System::getVerbose();

	System::printBannerOnce();

	DL_Loader *pLoader=new DL_Loader();
	FEASSERT(pLoader);

	BWORD success=FALSE;
	const BWORD adaptname=TRUE;

	String nameOrder=FE_STRING(FE_MUTEX_ORDER);
	String name;
	while(!nameOrder.empty())
	{
		while(!success && !(name=nameOrder.parse("\"",":")).empty())
		{
			if(name=="omp")
			{
#ifdef _OPENMP
				if(verbose=="all")
				{
					printf("%susing OpenMP (for mutexes)\n",
							PrefixLog::prefix().c_str());
				}
				ms_support=name;

				delete pLoader;
				return omp_mutex_init();
#endif
			}

			success=pLoader->openDL(name,adaptname);
		}

		if(!success)
		{
			if(verbose!="none")
			{
				printf("%sno mutex support\n",
						PrefixLog::prefix().c_str());
			}
			delete pLoader;
			return FALSE;
		}

		bool (*mutex_init)(void);
		mutex_init=(bool (*)(void))pLoader->findDLFunction("mutex_init");

		if(!mutex_init)
		{
			if(verbose!="none")
			{
				fprintf(stderr,"%smutex_init not found for %s\n",
						PrefixLog::prefix().c_str(),name.c_str());
			}
		}
		else if(mutex_init())
		{
			if(verbose=="all")
			{
				printf("%susing %s (for mutexes)\n",
						PrefixLog::prefix().c_str(),name.c_str());
			}

			ms_support=name;
			return TRUE;
		}
		else
		{
#if FE_THREAD_DEBUG
			fprintf(stderr,"%smutex_init failed for %s\n",
					PrefixLog::prefix().c_str(),name.c_str());
#endif
		}

		success=FALSE;
	}

	delete pLoader;
	return FALSE;

	//* TODO retain, close, and delete pLoader later
}

} /* namespace */

